"""Numerical approximation methods"""

from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import ulab

def interp(
    x: ulab.numpy.ndarray,
    xp: ulab.numpy.ndarray,
    fp: ulab.numpy.ndarray,
    *,
    left: Optional[_float] = None,
    right: Optional[_float] = None
) -> ulab.numpy.ndarray:
    """
    :param ulab.numpy.ndarray x: The x-coordinates at which to evaluate the interpolated values.
    :param ulab.numpy.ndarray xp: The x-coordinates of the data points, must be increasing
    :param ulab.numpy.ndarray fp: The y-coordinates of the data points, same length as xp
    :param left: Value to return for ``x < xp[0]``, default is ``fp[0]``.
    :param right: Value to return for ``x > xp[-1]``, default is ``fp[-1]``.

    Returns the one-dimensional piecewise linear interpolant to a function with given discrete data points (xp, fp), evaluated at x."""
    ...

def trapz(
    y: ulab.numpy.ndarray, x: Optional[ulab.numpy.ndarray] = None, dx: _float = 1.0
) -> _float:
    """
    :param 1D ulab.numpy.ndarray y: the values of the dependent variable
    :param 1D ulab.numpy.ndarray x: optional, the coordinates of the independent variable. Defaults to uniformly spaced values.
    :param float dx: the spacing between sample points, if x=None

    Returns the integral of y(x) using the trapezoidal rule.
    """
    ...

"""Numerical and Statistical functions

Most of these functions take an "axis" argument, which indicates whether to
operate over the flattened array (None), or a particular axis (integer)."""

from typing import Dict

_ArrayLike = Union[ndarray, List[_float], Tuple[_float], range]

_DType = int
"""`ulab.numpy.int8`, `ulab.numpy.uint8`, `ulab.numpy.int16`, `ulab.numpy.uint16`, `ulab.numpy.float` or `ulab.numpy.bool`"""

_float = float
"""Type alias of the bulitin float"""

_bool = bool
"""Type alias of the bulitin bool"""

int8: _DType
"""Type code for signed integers in the range -128 .. 127 inclusive, like the 'b' typecode of `array.array`"""

int16: _DType
"""Type code for signed integers in the range -32768 .. 32767 inclusive, like the 'h' typecode of `array.array`"""

float: _DType
"""Type code for floating point values, like the 'f' typecode of `array.array`"""

uint8: _DType
"""Type code for unsigned integers in the range 0 .. 255 inclusive, like the 'H' typecode of `array.array`"""

uint16: _DType
"""Type code for unsigned integers in the range 0 .. 65535 inclusive, like the 'h' typecode of `array.array`"""

bool: _DType
"""Type code for boolean values"""

def argmax(array: _ArrayLike, *, axis: Optional[int] = None) -> int:
    """Return the index of the maximum element of the 1D array"""
    ...

def argmin(array: _ArrayLike, *, axis: Optional[int] = None) -> int:
    """Return the index of the minimum element of the 1D array"""
    ...

def argsort(array: ulab.numpy.ndarray, *, axis: int = -1) -> ulab.numpy.ndarray:
    """Returns an array which gives indices into the input array from least to greatest."""
    ...

def cross(a: ulab.numpy.ndarray, b: ulab.numpy.ndarray) -> ulab.numpy.ndarray:
    """Return the cross product of two vectors of length 3"""
    ...

def diff(
    array: ulab.numpy.ndarray, *, n: int = 1, axis: int = -1
) -> ulab.numpy.ndarray:
    """Return the numerical derivative of successive elements of the array, as
    an array.  axis=None is not supported."""
    ...

def flip(
    array: ulab.numpy.ndarray, *, axis: Optional[int] = None
) -> ulab.numpy.ndarray:
    """Returns a new array that reverses the order of the elements along the
    given axis, or along all axes if axis is None."""
    ...

def max(array: _ArrayLike, *, axis: Optional[int] = None) -> _float:
    """Return the maximum element of the 1D array"""
    ...

def mean(array: _ArrayLike, *, axis: Optional[int] = None) -> _float:
    """Return the mean element of the 1D array, as a number if axis is None, otherwise as an array."""
    ...

def median(array: ulab.numpy.ndarray, *, axis: int = -1) -> ulab.numpy.ndarray:
    """Find the median value in an array along the given axis, or along all axes if axis is None."""
    ...

def min(array: _ArrayLike, *, axis: Optional[int] = None) -> _float:
    """Return the minimum element of the 1D array"""
    ...

def roll(
    array: ulab.numpy.ndarray, distance: int, *, axis: Optional[int] = None
) -> None:
    """Shift the content of a vector by the positions given as the second
    argument. If the ``axis`` keyword is supplied, the shift is applied to
    the given axis.  The array is modified in place."""
    ...

def sort(array: ulab.numpy.ndarray, *, axis: int = -1) -> ulab.numpy.ndarray:
    """Sort the array along the given axis, or along all axes if axis is None.
    The array is modified in place."""
    ...

def std(array: _ArrayLike, *, axis: Optional[int] = None, ddof: int = 0) -> _float:
    """Return the standard deviation of the array, as a number if axis is None, otherwise as an array."""
    ...

def sum(
    array: _ArrayLike, *, axis: Optional[int] = None
) -> Union[_float, int, ulab.numpy.ndarray]:
    """Return the sum of the array, as a number if axis is None, otherwise as an array."""
    ...

"""Compatibility layer for numpy"""

class ndarray: ...

def get_printoptions() -> Dict[str, int]:
    """Get printing options"""
    ...

def set_printoptions(
    threshold: Optional[int] = None, edgeitems: Optional[int] = None
) -> None:
    """Set printing options"""
    ...

def ndinfo(array: ulab.numpy.ndarray) -> None: ...
def array(
    values: Union[ndarray, Iterable[Union[_float, _bool, Iterable[Any]]]],
    *,
    dtype: _DType = ulab.numpy.float
) -> ulab.numpy.ndarray:
    """alternate constructor function for `ulab.numpy.ndarray`. Mirrors numpy.array"""
    ...

def trace(m: ulab.numpy.ndarray) -> _float:
    """
    :param m: a square matrix

    Compute the trace of the matrix, the sum of its diagonal elements."""
    ...

def dot(
    m1: ulab.numpy.ndarray, m2: ulab.numpy.ndarray
) -> Union[ulab.numpy.ndarray, _float]:
    """
    :param ~ulab.numpy.ndarray m1: a matrix, or a vector
    :param ~ulab.numpy.ndarray m2: a matrix, or a vector

    Computes the product of two matrices, or two vectors. In the letter case, the inner product is returned."""
    ...

"""Element-by-element functions

These functions can operate on numbers, 1-D iterables, and arrays of 1 to 4 dimensions by
applying the function to every element in the array.  This is typically
much more efficient than expressing the same operation as a Python loop."""

def acos(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse cosine function"""
    ...

def acosh(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse hyperbolic cosine function"""
    ...

def asin(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse sine function"""
    ...

def asinh(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse hyperbolic sine function"""
    ...

def around(a: _ArrayLike, *, decimals: int = 0) -> ulab.numpy.ndarray:
    """Returns a new float array in which each element is rounded to
    ``decimals`` places."""
    ...

def atan(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse tangent function; the return values are in the
    range [-pi/2,pi/2]."""
    ...

def arctan2(ya: _ArrayLike, xa: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse tangent function of y/x; the return values are in
    the range [-pi, pi]."""
    ...

def atanh(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the inverse hyperbolic tangent function"""
    ...

def ceil(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Rounds numbers up to the next whole number"""
    ...

def cos(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the cosine function"""
    ...

def cosh(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the hyperbolic cosine function"""
    ...

def degrees(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Converts angles from radians to degrees"""
    ...

def erf(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the error function, which has applications in statistics"""
    ...

def erfc(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the complementary error function, which has applications in statistics"""
    ...

def exp(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the exponent function."""
    ...

def expm1(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes $e^x-1$.  In certain applications, using this function preserves numeric accuracy better than the `exp` function."""
    ...

def floor(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Rounds numbers up to the next whole number"""
    ...

def gamma(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the gamma function"""
    ...

def lgamma(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the natural log of the gamma function"""
    ...

def log(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the natural log"""
    ...

def log10(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the log base 10"""
    ...

def log2(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the log base 2"""
    ...

def radians(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Converts angles from degrees to radians"""
    ...

def sin(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the sine function"""
    ...

def sinh(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the hyperbolic sine"""
    ...

def sqrt(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the square root"""
    ...

def tan(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the tangent"""
    ...

def tanh(a: _ArrayLike) -> ulab.numpy.ndarray:
    """Computes the hyperbolic tangent"""
    ...

def vectorize(
    f: Union[Callable[[int], _float], Callable[[_float], _float]],
    *,
    otypes: Optional[_DType] = None
) -> Callable[[_ArrayLike], ulab.numpy.ndarray]:
    """
    :param callable f: The function to wrap
    :param otypes: List of array types that may be returned by the function.  None is interpreted to mean the return value is float.

    Wrap a Python function ``f`` so that it can be applied to arrays.
    The callable must return only values of the types specified by ``otypes``, or the result is undefined."""
    ...



"""Support for PS/2 protocol

The `ps2io` module contains classes to provide PS/2 communication.

.. warning:: This module is not available in some SAMD21 builds. See the
  :ref:`module-support-matrix` for more info.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info."""

from __future__ import annotations

import microcontroller

class Ps2:
    """Communicate with a PS/2 keyboard or mouse

    Ps2 implements the PS/2 keyboard/mouse serial protocol, used in
    legacy devices. It is similar to UART but there are only two
    lines (Data and Clock). PS/2 devices are 5V, so bidirectional
    level converters must be used to connect the I/O lines to pins
    of 3.3V boards."""

    def __init__(
        self, data_pin: microcontroller.Pin, clock_pin: microcontroller.Pin
    ) -> None:
        """Create a Ps2 object associated with the given pins.

        :param ~microcontroller.Pin data_pin: Pin tied to data wire.
        :param ~microcontroller.Pin clock_pin: Pin tied to clock wire.
          This pin must support interrupts.

        Read one byte from PS/2 keyboard and turn on Scroll Lock LED::

          import ps2io
          import board

          kbd = ps2io.Ps2(board.D10, board.D11)

          while len(kbd) == 0:
              pass

          print(kbd.popleft())
          print(kbd.sendcmd(0xed))
          print(kbd.sendcmd(0x01))"""
        ...
    def deinit(self) -> None:
        """Deinitialises the Ps2 and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> Ps2:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def popleft(self) -> int:
        """Removes and returns the oldest received byte. When buffer
        is empty, raises an IndexError exception."""
        ...
    def sendcmd(self, byte: int) -> int:
        """Sends a command byte to PS/2. Returns the response byte, typically
        the general ack value (0xFA). Some commands return additional data
        which is available through :py:func:`popleft()`.

        Raises a RuntimeError in case of failure. The root cause can be found
        by calling :py:func:`clear_errors()`. It is advisable to call
        :py:func:`clear_errors()` before :py:func:`sendcmd()` to flush any
        previous errors.

        :param int byte: byte value of the command"""
        ...
    def clear_errors(self) -> None:
        """Returns and clears a bitmap with latest recorded communication errors.

        Reception errors (arise asynchronously, as data is received):

        0x01: start bit not 0

        0x02: timeout

        0x04: parity bit error

        0x08: stop bit not 1

        0x10: buffer overflow, newest data discarded

        Transmission errors (can only arise in the course of sendcmd()):

        0x100: clock pin didn't go to LO in time

        0x200: clock pin didn't go to HI in time

        0x400: data pin didn't ACK

        0x800: clock pin didn't ACK

        0x1000: device didn't respond to RTS

        0x2000: device didn't send a response byte in time"""
        ...
    def __bool__(self) -> bool: ...
    def __len__(self) -> int:
        """Returns the number of received bytes in buffer, available
        to :py:func:`popleft()`."""
        ...



"""Manipulate numeric data similar to numpy

`ulab` is a numpy-like module for micropython, meant to simplify and
speed up common mathematical operations on arrays. The primary goal was to
implement a small subset of numpy that might be useful in the context of a
microcontroller. This means low-level data processing of linear (array) and
two-dimensional (matrix) data.

`ulab` is adapted from micropython-ulab, and the original project's
documentation can be found at
https://micropython-ulab.readthedocs.io/en/latest/

`ulab` is modeled after numpy, and aims to be a compatible subset where
possible.  Numpy's documentation can be found at
https://docs.scipy.org/doc/numpy/index.html"""

from __future__ import annotations

from typing import Union

from .numpy import ndarray as _ndarray

ndarray = Union[_ndarray]
"""A temporary alias for the type `ulab.numpy.ndarray` needed during a transitional period. Do not use."""



"""Hardware accelerated external bus access

The I2CDevice and SPIDevice helper classes make managing transaction state on a bus easy.
For example, they manage locking the bus to prevent other concurrent access. For SPI
devices, it manages the chip select and protocol changes such as mode. For I2C, it
manages the device address."""

from __future__ import annotations

import busio
import microcontroller
from _typing import ReadableBuffer, WriteableBuffer

class I2CDevice:
    """I2C Device Manager"""

    def __init__(self, i2c: busio.I2C, device_address: int, probe: bool = True) -> None:

        """Represents a single I2C device and manages locking the bus and the device
        address.

        :param ~busio.I2C i2c: The I2C bus the device is on
        :param int device_address: The 7 bit device address
        :param bool probe: Probe for the device upon object creation, default is true

        Example::

            import busio
            from board import *
            from adafruit_bus_device.i2c_device import I2CDevice
            with busio.I2C(SCL, SDA) as i2c:
                device = I2CDevice(i2c, 0x70)
                bytes_read = bytearray(4)
                with device:
                    device.readinto(bytes_read)
                # A second transaction
                with device:
                    device.write(bytes_read)
        """
    ...
    def __enter__(self) -> I2CDevice:
        """Context manager entry to lock bus."""
        ...
    def __exit__(self) -> None:
        """Automatically unlocks the bus on exit."""
        ...
    import sys
    def readinto(
        self, buffer: WriteableBuffer, *, start: int = 0, end: int = sys.maxsize
    ) -> None:
        """Read into ``buffer`` from the device.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed.
        The number of bytes read will be the length of ``buffer[start:end]``.

        :param WriteableBuffer buffer: read bytes into this buffer
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``
        """
        ...
    import sys
    def write(
        self, buffer: ReadableBuffer, *, start: int = 0, end: int = sys.maxsize
    ) -> None:
        """Write the bytes from ``buffer`` to the device, then transmit a stop bit.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``buffer[start:end]``.

        :param ReadableBuffer buffer: write out bytes from this buffer
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``
        """
        ...
    import sys
    def write_then_readinto(
        self,
        out_buffer: ReadableBuffer,
        in_buffer: WriteableBuffer,
        *,
        out_start: int = 0,
        out_end: int = sys.maxsize,
        in_start: int = 0,
        in_end: int = sys.maxsize
    ) -> None:
        """Write the bytes from ``out_buffer`` to the device, then immediately
        reads into ``in_buffer`` from the device.

        If ``out_start`` or ``out_end`` is provided, then the buffer will be sliced
        as if ``out_buffer[out_start:out_end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``out_buffer[out_start:out_end]``.

        If ``in_start`` or ``in_end`` is provided, then the input buffer will be sliced
        as if ``in_buffer[in_start:in_end]`` were passed,
        The number of bytes read will be the length of ``out_buffer[in_start:in_end]``.

        :param ReadableBuffer out_buffer: write out bytes from this buffer
        :param WriteableBuffer in_buffer: read bytes into this buffer
        :param int out_start: beginning of ``out_buffer`` slice
        :param int out_end: end of ``out_buffer`` slice; if not specified, use ``len(out_buffer)``
        :param int in_start: beginning of ``in_buffer`` slice
        :param int in_end: end of ``in_buffer slice``; if not specified, use ``len(in_buffer)``
        """
        ...

class SPIDevice:
    """SPI Device Manager"""

    def __init__(
        self,
        spi: busio.SPI,
        chip_select: microcontroller.Pin,
        *,
        baudrate: int = 100000,
        polarity: int = 0,
        phase: int = 0,
        extra_clocks: int = 0
    ) -> None:

        """
        Represents a single SPI device and manages locking the bus and the device address.

        :param ~busio.SPI spi: The SPI bus the device is on
        :param ~digitalio.DigitalInOut chip_select: The chip select pin object that implements the DigitalInOut API.
        :param bool cs_active_value: Set to true if your device requires CS to be active high. Defaults to false.
        :param int extra_clocks: The minimum number of clock cycles to cycle the bus after CS is high. (Used for SD cards.)

        Example::

            import busio
            import digitalio
            from board import *
            from adafruit_bus_device.spi_device import SPIDevice
            with busio.SPI(SCK, MOSI, MISO) as spi_bus:
                cs = digitalio.DigitalInOut(D10)
                device = SPIDevice(spi_bus, cs)
                bytes_read = bytearray(4)
                # The object assigned to spi in the with statements below
                # is the original spi_bus object. We are using the busio.SPI
                # operations busio.SPI.readinto() and busio.SPI.write().
                with device as spi:
                    spi.readinto(bytes_read)
                # A second transaction
                with device as spi:
                    spi.write(bytes_read)"""
    ...
    def __enter__(self) -> busio.SPI:
        """Starts a SPI transaction by configuring the SPI and asserting chip select."""
        ...
    def __exit__(self) -> None:
        """Ends a SPI transaction by deasserting chip select. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...



"""Real Time Clock

The `rtc` module provides support for a Real Time Clock. You can access and manage the
RTC using :class:`rtc.RTC`. It also backs the :func:`time.time` and :func:`time.localtime`
functions using the onboard RTC if present."""

from __future__ import annotations

import time

def set_time_source(rtc: RTC) -> None:
    """Sets the RTC time source used by :func:`time.localtime`.
    The default is :class:`rtc.RTC`, but it's useful to use this to override the
    time source for testing purposes. For example::

      import rtc
      import time

      class RTC(object):
          @property
          def datetime(self):
              return time.struct_time((2018, 3, 17, 21, 1, 47, 0, 0, 0))

      r = RTC()
      rtc.set_time_source(r)"""
    ...

class RTC:
    """Real Time Clock"""

    def __init__(self) -> None:
        """This class represents the onboard Real Time Clock. It is a singleton and will always return the same instance."""
        ...
    datetime: time.struct_time
    """The current date and time of the RTC as a `time.struct_time`.

    This must be set to the current date and time whenever the board loses power::

      import rtc
      import time

      r = rtc.RTC()
      r.datetime = time.struct_time((2019, 5, 29, 15, 14, 15, 0, -1, -1))


    Once set, the RTC will automatically update this value as time passes. You can read this
    property to get a snapshot of the current time::

      current_time = r.datetime
      print(current_time)
      # struct_time(tm_year=2019, tm_month=5, ...)"""

    calibration: int
    """The RTC calibration value as an `int`.

    A positive value speeds up the clock and a negative value slows it down.
    Range and value is hardware specific, but one step is often approximately 1 ppm::

      import rtc
      import time

      r = rtc.RTC()
      r.calibration = 1"""



"""Getpass Module

This module provides a way to get input from user without echoing it.

"""

from __future__ import annotations

import io
from typing import Optional

...

def getpass(
    prompt: Optional[str] = "Password: ", stream: Optional[io.FileIO] = None
) -> str:

    """Prompt the user without echoing.

    :param str prompt: The user is prompted using the string ``prompt``, which defaults to ``'Password: '``.
    :param io.FileIO stream: The ``prompt`` is written to the file-like object ``stream`` if provided.

    """
    ...



"""Support for PWM based protocols

The `pwmio` module contains classes to provide access to basic pulse IO.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import time
  import pwmio
  import board

  pwm = pwmio.PWMOut(board.LED)
  pwm.duty_cycle = 2 ** 15
  time.sleep(0.1)

This example will initialize the the device, set
:py:data:`~pwmio.PWMOut.duty_cycle`, and then sleep 0.1 seconds.
CircuitPython will automatically turn off the PWM when it resets all
hardware after program completion. Use ``deinit()`` or a ``with`` statement
to do it yourself."""

from __future__ import annotations

import microcontroller

class PWMOut:
    """Output a Pulse Width Modulated signal on a given pin."""

    def __init__(
        self,
        pin: microcontroller.Pin,
        *,
        duty_cycle: int = 0,
        frequency: int = 500,
        variable_frequency: bool = False
    ) -> None:
        """Create a PWM object associated with the given pin. This allows you to
        write PWM signals out on the given pin. Frequency is fixed after init
        unless ``variable_frequency`` is True.

        .. note:: When ``variable_frequency`` is True, further PWM outputs may be
          limited because it may take more internal resources to be flexible. So,
          when outputting both fixed and flexible frequency signals construct the
          fixed outputs first.

        :param ~microcontroller.Pin pin: The pin to output to
        :param int duty_cycle: The fraction of each pulse which is high. 16-bit
        :param int frequency: The target frequency in Hertz (32-bit)
        :param bool variable_frequency: True if the frequency will change over time


        Simple LED on::

          import pwmio
          import board

          pwm = pwmio.PWMOut(board.LED)

          while True:
              pwm.duty_cycle = 2 ** 15  # Cycles the pin with 50% duty cycle (half of 2 ** 16) at the default 500hz

        PWM LED fade::

          import pwmio
          import board

          pwm = pwmio.PWMOut(board.LED)  # output on LED pin with default of 500Hz

          while True:
              for cycle in range(0, 65535):  # Cycles through the full PWM range from 0 to 65535
                  pwm.duty_cycle = cycle  # Cycles the LED pin duty cycle through the range of values
              for cycle in range(65534, 0, -1):  # Cycles through the PWM range backwards from 65534 to 0
                  pwm.duty_cycle = cycle  # Cycles the LED pin duty cycle through the range of values

        PWM at specific frequency (servos and motors)::

          import pwmio
          import board

          pwm = pwmio.PWMOut(board.D13, frequency=50)
          pwm.duty_cycle = 2 ** 15  # Cycles the pin with 50% duty cycle (half of 2 ** 16) at 50hz

        Variable frequency (usually tones)::

          import pwmio
          import board
          import time

          pwm = pwmio.PWMOut(board.D13, duty_cycle=2 ** 15, frequency=440, variable_frequency=True)
          time.sleep(0.2)
          pwm.frequency = 880
          time.sleep(0.1)"""
        ...
    def deinit(self) -> None:
        """Deinitialises the PWMOut and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> PWMOut:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    duty_cycle: int
    """16 bit value that dictates how much of one cycle is high (1) versus low
    (0). 0xffff will always be high, 0 will always be low and 0x7fff will
    be half high and then half low.

    Depending on how PWM is implemented on a specific board, the internal
    representation for duty cycle might have less than 16 bits of resolution.
    Reading this property will return the value from the internal representation,
    so it may differ from the value set."""

    frequency: int
    """32 bit value that dictates the PWM frequency in Hertz (cycles per
    second). Only writeable when constructed with ``variable_frequency=True``.

    Depending on how PWM is implemented on a specific board, the internal value
    for the PWM's duty cycle may need to be recalculated when the frequency
    changes. In these cases, the duty cycle is automatically recalculated
    from the original duty cycle value. This should happen without any need
    to manually re-set the duty cycle."""



"""A fast RGB(W) pixel buffer library for like NeoPixel and DotStar

The `adafruit_pixelbuf` module provides the :py:class:`PixelBuf` class to accelerate
RGB(W) strip/matrix manipulation, such as DotStar and Neopixel.

Also available as ``_pixelbuf``. This usage has been deprecated.

Byteorders are configured with strings, such as "RGB" or "RGBD"."""

from __future__ import annotations

from typing import List, Tuple, Union, overload

from _typing import ReadableBuffer

class PixelBuf:
    """A fast RGB[W] pixel buffer for LED and similar devices."""

    def __init__(
        self,
        size: int,
        *,
        byteorder: str = "BGR",
        brightness: float = 0,
        auto_write: bool = False,
        header: ReadableBuffer = b"",
        trailer: ReadableBuffer = b""
    ) -> None:
        """Create a PixelBuf object of the specified size, byteorder, and bits per pixel.

        When brightness is less than 1.0, a second buffer will be used to store the color values
        before they are adjusted for brightness.

        When ``P`` (PWM duration) is present as the 4th character of the byteorder
        string, the 4th value in the tuple/list for a pixel is the individual pixel
        brightness (0.0-1.0) and will enable a Dotstar compatible 1st byte for each
        pixel.

        :param int size: Number of pixels
        :param str byteorder: Byte order string (such as "RGB", "RGBW" or "PBGR")
        :param float brightness: Brightness (0 to 1.0, default 1.0)
        :param bool auto_write: Whether to automatically write pixels (Default False)
        :param ~_typing.ReadableBuffer header: Sequence of bytes to always send before pixel values.
        :param ~_typing.ReadableBuffer trailer: Sequence of bytes to always send after pixel values."""
        ...
    bpp: int
    """The number of bytes per pixel in the buffer (read-only)"""

    brightness: float
    """Float value between 0 and 1.  Output brightness.

    When brightness is less than 1.0, a second buffer will be used to store the color values
    before they are adjusted for brightness."""

    auto_write: bool
    """Whether to automatically write the pixels after each update."""

    byteorder: str
    """byteorder string for the buffer (read-only)"""
    def show(self) -> None:
        """Transmits the color data to the pixels so that they are shown. This is done automatically
        when `auto_write` is True."""
        ...
    def fill(
        self, color: Union[int, Tuple[int, int, int], Tuple[int, int, int, float]]
    ) -> None:
        """Fills the given pixelbuf with the given color."""
        ...
    @overload
    def __getitem__(
        self, index: slice
    ) -> Union[
        Tuple[Tuple[int, int, int], ...], Tuple[Tuple[int, int, int, float], ...]
    ]: ...
    @overload
    def __getitem__(
        self, index: int
    ) -> Union[Tuple[int, int, int], Tuple[int, int, int, float]]:
        """Returns the pixel value at the given index as a tuple of (Red, Green, Blue[, White]) values
        between 0 and 255.  When in PWM (DotStar) mode, the 4th tuple value is a float of the pixel
        intensity from 0-1.0."""
        ...
    @overload
    def __setitem__(
        self,
        index: slice,
        value: Tuple[Union[int, Tuple[float, ...], List[float]], ...],
    ) -> None: ...
    @overload
    def __setitem__(
        self, index: slice, value: List[Union[int, Tuple[float, ...], List[float]]]
    ) -> None: ...
    @overload
    def __setitem__(
        self, index: int, value: Union[int, Tuple[float, ...], List[float]]
    ) -> None:
        """Sets the pixel value at the given index.  Value can either be a tuple or integer.  Tuples are
        The individual (Red, Green, Blue[, White]) values between 0 and 255.  If given an integer, the
        red, green and blue values are packed into the lower three bytes (0xRRGGBB).
        For RGBW byteorders, if given only RGB values either as an int or as a tuple, the white value
        is used instead when the red, green, and blue values are the same."""
        ...
